# cython: embedsignature=True


import pyximport
pyximport.install()


cimport cpython.ref as cpy_ref
#from cpython.ref cimport PyObject
from cpython cimport PyObject, Py_INCREF

from cylp.cy.CyClpPrimalColumnPivotBase cimport CyClpPrimalColumnPivotBase
from cylp.cy.CyClpDualRowPivotBase cimport CyClpDualRowPivotBase
#from cylp.cy.CyCoinIndexedVector cimport CyCoinIndexedVector, CppCoinIndexedVector
from cylp.cy.CyCoinModel cimport CyCoinModel, CppCoinModel
from cylp.cy.CyCoinPackedMatrix cimport CyCoinPackedMatrix, CppCoinPackedMatrix
from cylp.cy.CyCbcModel cimport CyCbcModel, CppICbcModel
from cylp.python.modeling.CyLPModel import CyLPModel
from cylp.cy.CyCoinIndexedVector cimport CyCoinIndexedVector, CppCoinIndexedVector

from libcpp.string cimport string
from libcpp.vector cimport vector

cdef extern from "IClpPrimalColumnPivotBase.h" namespace "IClpSimplex":
    cdef enum Status:
        isFree = 0x00,
        basic = 0x01,
        atUpperBound = 0x02,
        atLowerBound = 0x03,
        superBasic = 0x04,
        isFixed = 0x05

cdef extern from "ClpPrimalColumnPivot.hpp":
    cdef cppclass CppClpPrimalColumnPivot "ClpPrimalColumnPivot":
        pass
    CppClpPrimalColumnPivot *new_ClpPrimalColumnPivot \
                                    "new ClpPrimalColumnPivot" ()

cdef extern from "ClpDualRowPivot.hpp":
    cdef cppclass CppClpDualRowPivot "ClpDualRowPivot":
        pass
    CppClpDualRowPivot *new_ClpDualRowPivot \
                                    "new ClpDualRowPivot" ()

cdef extern from "IClpSimplex.hpp":

    ctypedef int (*runIsPivotAcceptable_t)(void* obj)
    ctypedef int (*varSelCriteria_t)(void* obj, int varInd)
    cdef double cdot(CppCoinIndexedVector* pv1, CppCoinIndexedVector* pv2)

    cdef cppclass CppIClpSimplex "IClpSimplex":
        CppIClpSimplex(PyObject* obj,
                       runIsPivotAcceptable_t runIsPivotAcceptable,
                       varSelCriteria_t runVarSelCriteria)

        void setInteger(int index)
        void copyInIntegerInformation(char* information)

        void setCriteria(varSelCriteria_t vsc)
        void setPrimalColumnPivotAlgorithm(CppClpPrimalColumnPivot* choice)
        void setDualRowPivotAlgorithm(CppClpDualRowPivot* choice)
        int readMps(char*, int keepNames, int ignoreErrors)
        void loadQuadraticObjective(CppCoinPackedMatrix* matrix)
        CppCoinPackedMatrix* getMatrix()
        int primal(int ifValuesPass, int startFinishOptions)
        int dual(int ifValuesPass, int startFinishOptions)
        int initialSolve(int presolveType)
        int initialPrimalSolve()
        int initialDualSolve()
        void setPerturbation(int value)
        double* djRegion()
        int getNumCols()
        int getNumRows()
        Status getStatus(int sequence)
        void setStatus(int sequence, Status newstatus)
        double objectiveValue()
        int numberIterations()
        int* QP_ComplementarityList
        int* QP_BanList
        int QP_ExistsBannedVariable

        void useCustomPrimal(int customPrimal)
        int getUseCustomPrimal()

        void setObjectiveCoefficient(int elementIndex, double elementValue )
        void resize(int newNumberRows, int newNumberColumns)

        void setComplementarityList(int* cl)

        void addRow(int numberInRow,
                    int * columns,
                    double * elements,
                    double rowLower,
                    double rowUpper)

        void addColumn(int numberInColumn,
                int * rows,
                double * elements,
                double columnLower,
                double  columnUpper,
                double  objective)

        #number is the number of columns to be added
        void addColumns(int number,
                        double * columnLower,
                        double * columnUpper,
                        double * objective,
                        int * columnStarts,
                        int * rows,
                        double * elements)

        void deleteColumns(int number, int * which)
        void deleteRows(int number, int * which)

        #number is the number of rows to be added
        void addRows(int number,
                        double * rowLower,
                        double * rowUpper,
                        int * rowStarts,
                        int * columns,
                        double * elements)

        void getBasics(int* index)
        void getBInvACol(int col, double* vec)
        void getBInvCol(int col, double* vec)
        void getACol(int ncol, CppCoinIndexedVector * colArray)
        void getRightHandSide(double* righthandside)

        void setColumnUpper(int elementIndex, double elementValue)
        void setColumnLower(int elementIndex, double elementValue)
        void setRowUpper(int elementIndex, double elementValue)
        void setRowLower(int elementIndex, double elementValue)

        double* primalColumnSolution()
        double* dualColumnSolution()
        double* primalRowSolution()
        double* dualRowSolution()
        int status()

        bint flagged(int sequence)
        void setFlagged(int varInd)

        double largestDualError()

        int pivotRow()
        void setPivotRow(int v)

        int sequenceIn()
        void setSequenceIn(int v)

        double currentDualTolerance()
        double dualTolerance()
        void setDualTolerance(double value)
        double primalTolerance()
        void setPrimalTolerance(double value)

        double* rowUpper()
        double* rowLower()
        int numberRows()
        int* ComplementarityList()
        int * pivotVariable()
        #void computeDuals()

        #methods that return nunmpy arrays from c (double*  ,...)
        PyObject* getReducedCosts()
        void setReducedCosts(double* rc)
        PyObject* getStatusArray()
        PyObject* getComplementarityList()
        PyObject* getPivotVariable()

        PyObject* getPrimalRowSolution()
        PyObject* getPrimalColumnSolution()
        PyObject* getPrimalColumnSolutionAll()
        PyObject* getSolutionRegion()
        PyObject* getCostRegion()
        PyObject* getDualRowSolution()
        PyObject* getDualColumnSolution()

        PyObject* filterVars(PyObject*)

        void vectorTimesB_1(CppCoinIndexedVector* vec)
        void transposeTimesSubset(int number, int* which,
                                  double* pi, double* y)
        void transposeTimes(CppIClpSimplex * model, double scalar,
                                 CppCoinIndexedVector * x,
                                 CppCoinIndexedVector * y,
                                 CppCoinIndexedVector * z)
        void transposeTimesSubsetAll(int number, long long int* which,
                                     double* pi, double* y)
        int updateColumnFT(CppCoinIndexedVector* spare, CppCoinIndexedVector* updatedColumn)
        int updateColumnTranspose(CppCoinIndexedVector* regionSparse1,
                                  CppCoinIndexedVector* regionSparse2)

        CppIClpSimplex* preSolve(CppIClpSimplex* si,
                              double feasibilityTolerance,
                              bint keepIntegers,
                              int numberPasses,
                              bint dropNames,
                              bint doRowObjective)
        void postSolve(bint updateStatus)
        int dualWithPresolve(CppIClpSimplex* si,
                              double feasibilityTolerance,
                              bint keepIntegers,
                              int numberPasses,
                              bint dropNames,
                              bint doRowObjective)
        int primalWithPresolve(CppIClpSimplex* si,
                              double feasibilityTolerance,
                              bint keepIntegers,
                              int numberPasses,
                              bint dropNames,
                              bint doRowObjective)

        int loadProblem(CppCoinModel * modelObject, int tryPlusMinusOne)
        void loadProblem(CppCoinPackedMatrix* matrix,
		                  double* collb,  double* colub,
		                  double* obj,
		                  double* rowlb,  double* rowub,
		                  double * rowObjective)

        void setComplement(int var1, int var2)

        void replaceMatrix(CppCoinPackedMatrix* matrix, bint deleteCurrent)

        double getCoinInfinity()

        void setColumnUpperArray(double* columnUpper)
        void setColumnLowerArray(double* columnLower)
        void setColumnUpperSubset(int n, int* indOfind, int *indices, double* values)
        void setColumnLowerSubset(int n, int* indOfind, int *indices, double* values)
        void setColumnUpperFirstElements(int n, double* values)
        void setColumnLowerFirstElements(int n, double* values)
        void setRowUpperArray(double* rowUpper)
        void setRowLowerArray(double* rowLower)
        void setObjectiveArray(double* objective, int numberColumns)
        PyObject* getRowLower()
        PyObject* getRowUpper()
        PyObject* getLower()
        PyObject* getUpper()
        PyObject* getColLower()
        PyObject* getColUpper()
        PyObject* getObjective()
        PyObject* getColumnScale()
        PyObject* getRowScale()
        int writeMps(char* filename, int formatType, int numberAcross,
                     double objSense)
        int readLp(char *filename, double epsilon)

        void setVariableName(int varInd, char* name)
        void setConstraintName(int constInd, char* name)
        vector[string] getVariableNames()

        int partialPrice(int start, int end, int* numberWanted)

        int varIsFree(int ind)
        int varBasic(int ind)
        int varAtUpperBound(int ind)
        int varAtLowerBound(int ind)
        int varSuperBasic(int ind)
        int varIsFixed(int ind)

        #int argWeightedMax(PyObject* arr, PyObject* whr, double weight)
        int argWeightedMax(PyObject* arr, PyObject* arr_ind, PyObject* w,
                            PyObject* w_ind)

        CppICbcModel* getICbcModel()
        void writeLp(char *filename, char *extension,
                       double epsilon, int numberAcross,
                       int decimals, double objSense,
                       bint useRowNames)

        void setMaxNumIteration(int m)
        int maximumIterations()

        #Osi
        void setBasisStatus(int* cstat, int* rstat)
        void getBasisStatus(int* cstat, int* rstat)

        void setLogLevel(int value)
        int logLevel()

        void setObjectiveOffset(double value)
        double objectiveOffset()

        bint automaticScaling()
        void setAutomaticScaling(bint value)

        void scaling(int mode)
        int scalingFlag()

        void setInfeasibilityCost(double value)
        double infeasibilityCost()

        int numberPrimalInfeasibilities()

        bint isInteger(int index)
        PyObject* getIntegerInformation()

        double optimizationDirection()
        void setOptimizationDirection(double value)

cdef class CyClpSimplex:
    '''
    This is the documentation of CyClpSimpelx in the pyx class
    '''

    cdef CppIClpSimplex *CppSelf
    cdef vars
    cdef object varSelCriteria
    cdef CyCoinModel coinModel
    cdef object cyLPModel
    cdef CyCbcModel cbcModel
    cdef object _Hessian

    #cdef void prepareForCython(self, int useCustomPrimal)
    cdef setCppSelf(self,  CppIClpSimplex* s)

    cdef CyClpPrimalColumnPivotBase cyPivot
    cdef CyClpDualRowPivotBase cyDualPivot
    #cdef CppICbcModel* cbcModel
    #cdef object nodeCompareObject
    #cdef cbcModelExists
    #cdef object pivotMethodObject
    #cdef object isPivotAcceptable_func

    cpdef int readMps(self, filename, int keepNames=*,
                int ignoreErrors=*) except *

    cdef setPrimalColumnPivotAlgorithm(self, void* choice)
    cdef setDualRowPivotAlgorithm(self, void* choice)
    cdef double* primalColumnSolution(self)
    cdef double* dualColumnSolution(self)
    cdef double* primalRowSolution(self)
    cdef double* dualRowSolution(self)
    cdef double* rowLower(self)
    cdef double* rowUpper(self)

    #methods that return numpy arrays from c (double*  ,...)
    cpdef getReducedCosts(self)
    cpdef getStatusArray(self)
    cpdef getComplementarityList(self)
    cpdef getPivotVariable(self)

    cpdef filterVars(self, inds)

    cpdef CLP_getVarStatus(self, int sequence)
    cpdef CLP_setVarStatus(self, int sequence, int status)

    cdef primalRow(self, CppCoinIndexedVector*,
                                CppCoinIndexedVector*,
                                CppCoinIndexedVector*,
                                CppCoinIndexedVector*,
                                int)

    #cdef void CLP_getBInvACol(self, int col, double* vec)
    #cdef void CLP_getRightHandSide(self, double* righthandside)

    cpdef getACol(self, int ncol, CyCoinIndexedVector colArray)

    #cdef void CLP_setComplementarityList(self, int*)
    cdef int* ComplementarityList(self)
    cdef int* pivotVariable(self)

    cpdef vectorTimesB_1(self, CyCoinIndexedVector vec)

    #cpdef int loadProblem(self, CyCoinModel modelObject, int tryPlusMinusOne=*)

    #cpdef getPrimalConstraintSolution(self)
    #cpdef getPrimalVariableSolution(self)
    #cpdef getDualConstraintSolution(self)
    #cpdef getDualVariableSolution(self)
    #cpdef createComplementarityList(self)

    cpdef setVariableName(self, varInd, name)
    cpdef setConstraintName(self, constInd, name)

cdef class VarStatus:
    pass
cpdef cydot(CyCoinIndexedVector v1, CyCoinIndexedVector v2)




from __future__ import print_function
#import random
import numpy as np
from numpy import random
from scipy import sparse
from cylp.cy import CyClpSimplex
from cylp.py.modeling import CyLPArray
from cylp.py.utils.sparseUtil import csr_matrixPlus
from cylp.py.QP.QP import QP

def getA(nRows, nCols, nnzPerCol):
    '''Return a sparse coef. matrix of a set-partitioning problem
    *nnzPerCol* specifies the number of non-zero elements in
    each column.
    '''
    A = sparse.lil_matrix((nRows, nCols))
    solCols = []
    for nRow in range(nRows):
        nCol = random.randint(0, nCols-1)
        A[nRow, nCol] = 1
        solCols.append(nCol)

    for nCol in [j for j in range(nCols) if j not in solCols]:
        for i in range(nnzPerCol):
            #if random.randint(0, 1):
                A[random.randint(0, nRows-1), nCol] = 1
            #else:
            #    A[random.randint(0, nRows-1), nCol] = -1

    return csr_matrixPlus(A)

def getG(nCols):
    n = nCols
    G = sparse.lil_matrix((n, n))
    for i in range(0, n-1):
        G[i, i] = 1
        #G[i+1, i] = -0.2
        #G[i, i+1] = -0.2
    G[nCols - 1, nCols - 1] = 1

    return csr_matrixPlus(G)

def getCoinInfinity():
    return 1.79769313486e+308

def generateRandomPositiveDefiniteMatrix(n, cond):
        random.seed()
        p = n
        N_of_P = np.matrix(np.random.standard_normal((p,p)))
        N_of_Q = np.matrix(np.random.standard_normal((n,n)))
        P = np.linalg.qr(N_of_P)[0]
        Q = np.linalg.qr(N_of_Q)[0]

        D = np.zeros((p, n))
        t_prime = cond**0.25
        D[0,0] = 1.0/t_prime
        D[n-1,n-1] = t_prime
        for i in range(1, n-1):
            D[i, i] = t_prime**(random.uniform(-1,1))

        C  = P * D * Q
        C = C / np.linalg.norm(C, 2)
        G = C.T * C
        G = 2 * G
        return G

def getQP(m, n):
    qp = QP()
    qp.G = csr_matrixPlus(generateRandomPositiveDefiniteMatrix(n, 10))  #getG(n)
    #qp.c = np.zeros(n)
    qp.c = np.random.random(n) * 50

    qp.C = getA(m, n, 25)
    qp.c_low = np.ones(m) * -getCoinInfinity()
    qp.c_up = np.ones(m)
#    qp.c_low = -np.ones(m)
#    qp.c_up = np.ones(m) * getCoinInfinity()
    #qp.c_up = (20 + np.random.random(m) * 40) * 10**-3
    qp.b = qp.A = 0
    qp.nInEquality = m

    #qp.C =qp.c_low = qp.c_up = 0
    #qp.A = getA(m, n, 17)
    #qp.b = np.ones(m)
    qp.nEquality = 0

    qp.x_low = np.ones(n) * -getCoinInfinity()
    qp.x_up = np.ones(n) * getCoinInfinity()

    qp.n = n
    qp.nOriginalVar = n
    qp.objectiveOffset = 0
    qp.filename = '/Users/mehdi/Desktop/qptest.txt'
    return qp

class QPGen:
    def __init__(self, n, m, HesCond, nnzPerCol, costRange, signs):
        self.n = n
        self.m = m
        self.HesCond = HesCond
        self.nnzPerCol = nnzPerCol
        self.costRange = costRange
        self.signs = signs
        self.generateQP()

    def generateQP(self):
        m = self.m
        n = self.n
        s = CyClpSimplex()

        iNonZero = set(random.randint(n, size=self.nnzPerCol))
        iZero = [i for i in range(n) if i not in iNonZero]
        x_star = np.matrix(np.zeros((n, 1)))
        z_star = np.matrix(np.zeros((n, 1)))

        for i in iNonZero:
            x_star[i, 0] = 1

        for i in iZero:
            z_star[i, 0] = 0 if random.randint(2) else random.random()

        G = getG(n)

        A = getA(m, n, self.nnzPerCol)
        y_star = np.matrix(random.random((m, 1)))

        c = -(G * x_star - A.T * y_star - z_star)

        obj = 0.5 * ((x_star.T * G) * x_star) + c.T * x_star
        print(obj)

        c = CyLPArray((c.T)[0])

        b = CyLPArray(((A * x_star).T)[0])
        b = np.squeeze(np.asarray(b))

        x = s.addVariable('x', n)

        s += A * x == b
        s += x >= 0


        c = CyLPArray(c)
        s.objective = c * x

        s.Hessian = G

        self.model = s
        return s

    def writeToFile(self):
        filename = 'qp_%d_%d_%d_%d_%d_%s.qps' % (self.n, self.m, self.HesCond,
                                   self.nnzPerCol, self.costRange, self.signs)
        filename = 'tests/' + filename
        print('writing to %s.' % filename)
        self.model.writeMps(filename)



if __name__ == '__main__':
    import cProfile
    import sys
    n, m = int(sys.argv[1]), int(sys.argv[2])
    print(m)
    print(m/10)
    qp = QPGen(n, m, 1, m/10, 10000, 'G')
    qp.writeToFile()




from .DantzigPivot import DantzigPivot
from .DualDantzigPivot import DualDantzigPivot
from .PositiveEdgePivot import PositiveEdgePivot
from .PivotPythonBase import PivotPythonBase
from .DualPivotPythonBase import DualPivotPythonBase
from .LIFOPivot import LIFOPivot
from .MostFrequentPivot import MostFrequentPivot
from .PositiveEdgeWolfePivot import PositiveEdgeWolfePivot
from .WolfePivot import WolfePivot



'''
This class defines an interface all the Python pivot rules must implement.
It consists of a method :py:func:`pivotColumn` which must be implemented and
:py:func:`isPivotAcceptable` whose implementation is optional.
'''

class PivotPythonBase:
    def pivotColumn(self):
        '''
        Every subclass of ``PivotPythonBase``, i.e. every pivot rule, must
        implement this method.

        Return the index of the entering variable, an integer.
        '''
        raise NotImplementedError('pivotColumn is not implemented')

    def isPivotAcceptable(self):
        '''
        This is run just before the actual pivoting happens. Return False
        if the pivot is not acceptable for your method.
        '''
        return True

    def updateReducedCosts(self, updates, spareRow1, spareRow2, spareCol1, spareCol2):
        '''
        Update the reduced costs as its done in CLP's Dantzig rule.
        '''
        s = self.clpModel
        if updates.nElements:
            s.updateColumnTranspose(spareRow2, updates)
            s.transposeTimes(-1, updates, spareCol2, spareCol1)
            s.reducedCosts[s.nVariables:][updates.indices] -= updates.elements[:updates.nElements]
            s.reducedCosts[:s.nVariables][spareCol1.indices] -= spareCol1.elements[:spareCol1.nElements]
        updates.clear()
        spareCol1.clear()




from cpython.ref cimport PyObject
cimport numpy as np
from cylp.cy.CyCoinPackedMatrix cimport CyCoinPackedMatrix, CppCoinPackedMatrix

cdef extern from "ICoinMpsIO.hpp":
    cdef cppclass CppICoinMpsIO "ICoinMpsIO":
        #PyObject* getIndicesNPArray()
        int readMps(char *filename)
        int getNumCols()
        int getNumRows()
        int getNumElements()
        PyObject* np_getColLower()
        PyObject* np_getColUpper()
        PyObject* np_getRowSense()
        PyObject* np_getRightHandSide()
        PyObject* np_getRowRange()
        PyObject* np_getRowLower()
        PyObject* np_getRowUpper()
        PyObject* np_getObjCoefficients()
        bint isContinuous(int colNumber)
        bint isInteger(int columnNumber)
        PyObject* np_integerColumns()
        char* rowName(int index)
        char* columnName(int index)
        int rowIndex(char * name)
        int columnIndex(char * name)
        char* getProblemName()

        CppCoinPackedMatrix * IgetMatrixByRow()
        CppCoinPackedMatrix * IgetMatrixByCol()

        PyObject* getQPColumnStarts()
        PyObject* getQPColumns()
        PyObject* getQPElements()

        double getObjectiveOffset()

        #int readQuadraticMps(char* filename,
        #       int * columnStart, int * column2, double * elements,
        #       int checkSymmetry)
        int IreadQuadraticMps(char* filename, int checkSymmetry)

    CppICoinMpsIO *new_CppICoinMpsIO "new CppICoinMpsIO" ()


cdef class CyCoinMpsIO:
    cdef CppICoinMpsIO *CppSelf
    cdef Hessian
#    cpdef  getColLower(self)
#    cpdef  getColUpper(self)
#    cpdef getRowSense(self)
#    cpdef getRightHandSide(self)
#    cpdef getRowRange(self)
#    cpdef getRowLower(self)
#    cpdef getRowUpper(self)
#    cpdef getObjCoefficients(self)
#    cpdef integerColumns(self)
#
#    cpdef getQPColumnStarts(self)
#    cpdef getQPColumns(self)
#    cpdef getQPElements(self)
#    cpdef getHessian(self)
#
#    cpdef getMatrixByRow(self)
#    cpdef getMatrixByCol(self)


    #cdef int CLP_readQuadraticMps(self, char* filename,
    #           int * columnStart, int * column2, double * elements,
    #           int checkSymmetry)




from cylp.cy.CyClpPrimalColumnPivotBase cimport *
cimport cython.operator.dereference as deref
from cylp.cy.CyClpSimplex cimport CyClpSimplex, CppIClpSimplex
from cylp.cy.CyCoinIndexedVector cimport CyCoinIndexedVector
from cpython cimport Py_INCREF, Py_DECREF

cdef class CyPivotPythonBase(CyClpPrimalColumnPivotBase):
	cdef pivotColumn(self, CppCoinIndexedVector* v1,
                     CppCoinIndexedVector* v2, CppCoinIndexedVector* v3,
					 CppCoinIndexedVector* v4, CppCoinIndexedVector* v5)
	cdef CyClpPrimalColumnPivot * clone(self, bint copyData)
	cdef void saveWeights(self, CppIClpSimplex * model, int mode)
	cdef object pivotMethodObject
